import re
import json

from .errors import MissingKeywordsInDictionary, MissingReplaceableLabels, DifferentReplaceableLabels


class Template:
    """
    This class generates a new reusable Templates.
    """
    _template_api = None
    _template_file = None
    _required_labels = None

    def __init__(self, path: str) -> None:
        """
        :param path str: Relative or Absolute path to folder containing files
                         workflow.json.template and workflow_api.json.template
                         Example: "./templates/sd15"
        :return: Tuple of strings - required labels to use template.
                 Example: ("CONFIG_NAME", "CHECKPOINT_NAME", "CLIP_SKIP")
        :raises MissingReplaceableLabels: One of template files completely missing replaceable labels.
                                          Reminder: Replaceable labels look like __CONFIG_NAME__.
        """
        with open(path + '/workflow_api.json.template', 'r', encoding='utf-8') as f:
            self._template_api = f.read()
            self._required_labels = tuple(re.findall("__(.+?)__", self._template_api))
            if len(self._required_labels) == 0:
                raise MissingReplaceableLabels('workflow_api.json.template')

        with open(path + '/workflow.json.template', 'r', encoding='utf-8') as f:
            self._template_file = f.read()
            temporaryLabels = tuple(re.findall("__(.+?)__", self._template_file))
            if len(temporaryLabels) == 0:
                raise MissingReplaceableLabels('workflow.json.template')

        for key in self._required_labels:
            if key not in temporaryLabels:
                raise DifferentReplaceableLabels(self._required_labels, temporaryLabels)
        for key in temporaryLabels:
            if key not in self._required_labels:
                raise DifferentReplaceableLabels(temporaryLabels, self._required_labels)

    def generate_for_api(self, replacements_dict: dict) -> dict:
        """
        Replaces every instance of __X__ to replacements_dict["X"] in api file.
        Loads this file as JSON object and returns it to be used in POST requests.

        :param replacements_dict dict: Dictonary with replacement words
                                       Example: {"X": "wow", "CONFIG_NAME": "interlace_sd15.ini", "CLIP_SKIP": "-1"}
        :return: Dictonary object (json)
        :raises MissingKeywordsInDictionary: Raised if one of keywords cannot be replaced,
                                             because it is missing in replacements_dict.
        """
        template = self._template_api
        for varname in self._required_labels:
            if varname not in replacements_dict:
                raise MissingKeywordsInDictionary(varname, self._required_labels)
            template = template.replace("__" + varname + "__", str(replacements_dict[varname]))

        return json.loads(template)

    def generate_workflow(self, replacements_dict: dict) -> dict:
        """
        Replaces every instance of __X__ to replacements_dict["X"] in loadable workflow file.
        Loads this file as JSON object and returns it to be able to be parsed.

        :param replacements_dict dict: Dictonary with replacement words
                                       Example: {"X": "wow", "CONFIG_NAME": "interlace_sd15.ini", "CLIP_SKIP": "-1"}
        :return: Dictonary object (json)
        :raises MissingKeywordsInDictionary: Raised if one of keywords cannot be replaced,
                                             because it is missing in replacements_dict.
        """
        template = self._template_file
        for varname in self._required_labels:
            if varname not in replacements_dict:
                raise MissingKeywordsInDictionary(varname, self._required_labels)
            template = template.replace("__" + varname + "__", str(replacements_dict[varname]))

        return json.loads(template)

    def generate_saveable_workflow(self, replacements_dict: dict) -> str:
        """
        Replaces every instance of __X__ to replacements_dict["X"] in loadable workflow file.
        Returns it as JSON-loadable string, so it can be saved into file.

        :param replacements_dict dict: Dictonary with replacement words
                                       Example: {"X": "wow", "CONFIG_NAME": "interlace_sd15.ini", "CLIP_SKIP": "-1"}
        :return: String (guaranteed to be JSON-parseable)
        :raises MissingKeywordsInDictionary: Raised if one of keywords cannot be replaced,
                                             because it is missing in replacements_dict.
        """
        return json.dumps(self.generate_workflow(replacements_dict))
